use std::ops::Range;

use range_ext::intersect::{Intersect, IntersectionExt};

/// Operations to apply to a vector
#[derive(Debug, Clone)]
pub enum Operation<T> {
    /// No operation
    Nop,
    /// Insert
    Ins { pos: usize, val: Vec<T> },
    /// Delete
    Del { pos: usize, n: usize },
    /// Move
    Mov { pos: usize, to: usize },
    /// Sequence of operations
    Seq { ops: Vec<Operation<T>> },
}

impl<T> Operation<T> {
    fn len(&self) -> usize {
        match self {
            Operation::Nop | Operation::Seq { .. } => 0,
            Operation::Ins { val, .. } => val.len(),
            Operation::Del { n, .. } => *n,
            Operation::Mov { .. } => 1,
        }
    }

    fn pos(&self) -> usize {
        match self {
            Operation::Nop | Operation::Seq { .. } => 0,
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => *pos,
        }
    }

    fn range(&self) -> Range<usize> {
        match self {
            Operation::Nop | Operation::Seq { .. } => 0..0,
            Operation::Ins { pos, val } => *pos..(*pos + val.len()),
            Operation::Del { pos, n } => *pos..(*pos + *n),
            Operation::Mov { pos, .. } => *pos..(*pos + 1),
        }
    }

    fn set_pos(&mut self, n_pos: usize) {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("cannot move nop/seq"),
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => {
                *pos = n_pos;
            }
        }
    }

    fn mov(&mut self, n: usize, dir: bool) {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("cannot move nop/seq"),
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => {
                if dir {
                    *pos += n;
                } else {
                    *pos -= n;
                }
            }
        }
    }

    pub fn apply(self, vec: &mut Vec<T>) {
        match self {
            Operation::Nop => {}
            Operation::Ins { pos, mut val } => {
                if pos < vec.len() {
                    vec.splice(pos..pos, val);
                } else {
                    vec.append(&mut val);
                }
            }
            Operation::Del { pos, n } => {
                vec.splice(pos..(pos + n), None);
            }
            Operation::Mov { pos, to } => {
                let buf = vec.remove(pos);
                vec.insert(to, buf);
            }
            Operation::Seq { ops } => ops.into_iter().for_each(|op| op.apply(vec)),
        }
    }
}

pub fn transform<T1, T2>(a: &mut Operation<T1>, b: &mut Operation<T2>) {
    match a {
        Operation::Nop => {}
        Operation::Ins { pos: p1, val: v1 } => match b {
            Operation::Nop => {}
            Operation::Ins {
                pos: p2, val: v2, ..
            } => {
                if p1 <= p2 {
                    *p2 += v1.len();
                } else {
                    *p1 += v2.len();
                }
            }
            Operation::Del { .. } => transform_ins_del(a, b),
            Operation::Mov { .. } => transform_mov(b, a, false),
            Operation::Seq { ops } => transform_seq(ops, a, false),
        },
        Operation::Del { pos: p1, n: n1 } => match b {
            Operation::Nop => {}
            Operation::Ins { .. } => transform_ins_del(b, a),
            Operation::Del { pos: p2, n: n2 } => {
                let r1 = *p1..*p1 + *n1;
                let r2 = *p2..*p2 + *n2;

                match r1.intersect_ext(&r2) {
                    IntersectionExt::Empty => {}
                    IntersectionExt::Less => {
                        *p2 -= *n1;
                    }
                    IntersectionExt::LessOverlap => {
                        let overlap = r1.end.saturating_sub(*p2);
                        *n2 -= overlap;
                        *p2 = *p2 + overlap - *n1;
                        *n1 -= overlap;
                    }
                    IntersectionExt::Within => {
                        *n2 -= *n1;
                        *a = Operation::Nop;
                    }
                    IntersectionExt::Same => {
                        *a = Operation::Nop;
                        *b = Operation::Nop;
                    }
                    IntersectionExt::Over => {
                        *n1 -= *n2;
                        *b = Operation::Nop;
                    }
                    IntersectionExt::GreaterOverlap => {
                        let overlap = r2.end.saturating_sub(*p1);
                        *n1 -= overlap;
                        *p1 = *p1 + overlap - *n2;
                        *n2 -= overlap;
                    }
                    IntersectionExt::Greater => {
                        *p1 -= *n2;
                    }
                }
            }
            Operation::Mov { .. } => transform_mov(b, a, false),
            Operation::Seq { ops } => transform_seq(ops, a, false),
        },
        Operation::Mov { .. } => transform_mov(a, b, true),
        Operation::Seq { ops } => transform_seq(ops, b, true),
    }
}

fn transform_ins_del<T1, T2>(ins: &mut Operation<T1>, del: &mut Operation<T2>) {
    let ilen = ins.len();
    let dlen = del.len();

    let ir = ins.range();
    let dr = del.range();

    if ir.start > dr.start && ir.start < dr.end {
        ins.set_pos(dr.start);
        *del = Operation::Seq {
            ops: vec![
                Operation::Del {
                    pos: ir.end,
                    n: dlen - (ir.start - dr.start),
                },
                Operation::Del {
                    pos: dr.start,
                    n: ir.start - dr.start,
                },
            ],
        };
    } else if ir.start <= dr.start {
        del.mov(ilen, true);
    } else {
        ins.set_pos(ir.start.checked_sub(dlen).unwrap_or_default())
    }
}

fn transform_seq<T1, T2>(a_ops: &mut Vec<Operation<T1>>, b: &mut Operation<T2>, first: bool) {
    if let Operation::Seq { ops } = b {
        ops.iter_mut()
            .for_each(|op| transform_seq(a_ops, op, first));
    } else {
        a_ops.iter_mut().for_each(|op| {
            if first {
                transform(op, b)
            } else {
                transform(b, op)
            }
        })
    }
}

fn transform_mov<T1, T2>(a: &mut Operation<T1>, b: &mut Operation<T2>, first: bool) {
    let ra = a.range();

    if let Operation::Mov {
        pos: m_pos,
        to: m_to,
    } = a
    {
        match b {
            Operation::Ins { .. } => {}
            Operation::Del { pos: d_pos, n } => {
                let rb = *d_pos..(*d_pos + *n);
                // Delete move source
                if ra == rb {
                    // Delete all moved items in A'
                    *d_pos = *m_to;
                    *a = Operation::Nop;
                    return;
                }
                // within src - within dst
                match (rb.contains(m_pos), rb.contains(m_to)) {
                    (true, false) => {
                        let p_rest = if m_to <= d_pos { *d_pos + 1 } else { *d_pos };
                        let mut ops = vec![
                            // Delete moved item
                            Operation::Del { pos: *m_to, n: 1 },
                            // Delete rest
                            Operation::Del {
                                pos: p_rest,
                                n: *n - 1,
                            },
                        ];
                        if *m_to < p_rest {
                            ops.swap(0, 1);
                        }

                        *a = Operation::Nop;
                        *b = Operation::Seq { ops };
                        return;
                    }
                    (false, true) => {
                        let mut p_start = *d_pos;
                        let to_r = m_to > m_pos;
                        let del_before_src = d_pos < m_pos;

                        if !del_before_src {
                            p_start -= 1;
                        }

                        // Number of items to delete before moved item
                        let n1 = *m_to - *d_pos + usize::from(to_r);

                        *m_to -= n1;
                        if del_before_src {
                            *m_pos -= *n;
                        }
                        if m_pos == m_to {
                            *a = Operation::Nop;
                        }

                        if n1 == *n {
                            *d_pos = p_start;
                        } else if n1 == 0 {
                            *d_pos = p_start + 1;
                        } else {
                            *b = Operation::Seq {
                                ops: vec![
                                    // Delete after moved item
                                    Operation::Del {
                                        pos: p_start + n1 + 1,
                                        n: *n - n1,
                                    },
                                    // Delete before moved item
                                    Operation::Del {
                                        pos: p_start,
                                        n: n1,
                                    },
                                ],
                            };
                        }
                        return;
                    }
                    (true, true) => {
                        // Within src and dst: delete whole range
                        *a = Operation::Nop;
                        return;
                    }
                    (false, false) => {}
                }
            }
            Operation::Mov {
                pos: b_pos,
                to: b_to,
            } => {
                if m_pos == b_pos {
                    if first {
                        if b_to == m_to {
                            *b = Operation::Nop;
                        } else {
                            *b_pos = *m_to;
                        }
                        *a = Operation::Nop;
                    } else {
                        if m_to == b_to {
                            *a = Operation::Nop;
                        } else {
                            *m_pos = *b_to;
                        }
                        *b = Operation::Nop;
                    }
                    return;
                }
            }
            Operation::Seq { ops } => return transform_seq(ops, a, false),
            Operation::Nop => return,
        }

        let mut del = Operation::Del::<()> { pos: *m_pos, n: 1 };
        let mut ins = Operation::Ins::<()> {
            pos: *m_to,
            val: vec![()],
        };

        if first {
            transform(&mut del, b);
            transform(&mut ins, b);
        } else {
            transform(b, &mut del);
            transform(b, &mut ins);
        }

        *m_pos = del.pos();
        *m_to = ins.pos();
        if m_pos == m_to {
            *a = Operation::Nop;
        }
    }
}
