mod util;

use proptest::prelude::*;

use otvec::Operation;
use util::{test_transform_cond, transformed_mov};

const ALL_SIZE: usize = 20;
const VEC_SIZE: usize = 100;

fn testvec(len: usize) -> Vec<usize> {
    (0..len).collect::<Vec<_>>()
}

fn filter_to(size: usize, pos: usize, to: usize) -> usize {
    if pos == to {
        if to >= size - 1 {
            filter_to(size, pos, 0)
        } else {
            filter_to(size, pos, to + 1)
        }
    } else {
        to
    }
}

fn t_ins_ins(size: usize, a_pos: usize, a_len: usize, b_pos: usize, b_len: usize) {
    let input = testvec(size);
    let a = Operation::Ins {
        pos: a_pos,
        val: testvec(a_len),
    };
    let b = Operation::Ins {
        pos: b_pos,
        val: testvec(b_len),
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

fn t_del_del(size: usize, a_pos: usize, a_len: usize, b_pos: usize, b_len: usize) {
    let input = testvec(size);
    let a = Operation::Del {
        pos: a_pos,
        n: a_len,
    };
    let b = Operation::Del {
        pos: b_pos,
        n: b_len,
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

fn t_ins_del(size: usize, a_pos: usize, a_len: usize, b_pos: usize, b_len: usize) {
    let input = testvec(size);

    let a = Operation::Ins {
        pos: a_pos,
        val: testvec(a_len),
    };
    let b = Operation::Del {
        pos: b_pos,
        n: b_len,
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

fn t_mov_ins(size: usize, a_pos: usize, a_len: usize, a_to: usize, b_pos: usize, b_len: usize) {
    let input = testvec(size);
    let a = transformed_mov(a_pos, a_len, a_to);
    let b = Operation::Ins {
        pos: b_pos,
        val: testvec(b_len),
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

fn t_mov_del(size: usize, a_pos: usize, a_len: usize, a_to: usize, b_pos: usize, b_len: usize) {
    let input = testvec(size);
    let a = transformed_mov(a_pos, a_len, a_to);
    let b = Operation::Del {
        pos: b_pos,
        n: b_len,
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

fn t_mov_mov(size: usize, a_pos: usize, a_to: usize, b_pos: usize, b_to: usize) {
    let input = testvec(size);
    let a = Operation::Mov {
        pos: a_pos,
        to: a_to,
    };
    let b = Operation::Mov {
        pos: b_pos,
        to: b_to,
    };
    test_transform_cond(&a, &b, &input);
    test_transform_cond(&b, &a, &input);
}

#[test]
fn all_ins_ins() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_len in 1..ALL_SIZE {
            for b_pos in 0..ALL_SIZE {
                for b_len in 1..ALL_SIZE {
                    t_ins_ins(ALL_SIZE, a_pos, a_len, b_pos, b_len)
                }
            }
        }
    }
}

#[test]
fn all_del_del() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_len in 1..(ALL_SIZE - a_pos) {
            for b_pos in 0..ALL_SIZE {
                for b_len in 1..(ALL_SIZE - b_pos) {
                    t_del_del(ALL_SIZE, a_pos, a_len, b_pos, b_len)
                }
            }
        }
    }
}

#[test]
fn all_ins_del() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_len in 1..ALL_SIZE {
            for b_pos in 0..ALL_SIZE {
                for b_len in 1..(ALL_SIZE - b_pos) {
                    t_ins_del(ALL_SIZE, a_pos, a_len, b_pos, b_len)
                }
            }
        }
    }
}

#[test]
fn all_mov_ins() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_len in 1..(ALL_SIZE - a_pos) {
            for a_to in 0..ALL_SIZE {
                let a_to = filter_to(ALL_SIZE, a_pos, a_to);
                let a_len = a_len.min(ALL_SIZE - a_pos.max(a_to));

                for b_pos in 0..ALL_SIZE {
                    for b_len in 1..ALL_SIZE {
                        t_mov_ins(ALL_SIZE, a_pos, a_len, a_to, b_pos, b_len)
                    }
                }
            }
        }
    }
}

#[test]
fn all_mov_del() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_len in 1..(ALL_SIZE - a_pos) {
            for a_to in 0..ALL_SIZE {
                let a_to = filter_to(ALL_SIZE, a_pos, a_to);
                let a_len = a_len.min(ALL_SIZE - a_pos.max(a_to));

                for b_pos in 0..ALL_SIZE {
                    for b_len in 1..(ALL_SIZE - b_pos) {
                        t_mov_del(ALL_SIZE, a_pos, a_len, a_to, b_pos, b_len)
                    }
                }
            }
        }
    }
}

#[test]
fn all_mov_mov() {
    for a_pos in 0..ALL_SIZE {
        println!("{a_pos}");
        for a_to in 0..ALL_SIZE {
            for b_pos in 0..ALL_SIZE {
                for b_to in 0..ALL_SIZE {
                    t_mov_mov(ALL_SIZE, a_pos, a_to, b_pos, b_to)
                }
            }
        }
    }
}

proptest! {
    #[test]
    fn transform_ins_ins(a_pos in 0usize..VEC_SIZE, a_len in 1usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_len in 1usize..VEC_SIZE) {
        t_ins_ins(VEC_SIZE, a_pos, a_len, b_pos, b_len)
    }

    #[test]
    fn transform_del_del(a_pos in 0usize..VEC_SIZE, a_len in 1usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_len in 1usize..VEC_SIZE) {
        // Limit inputs
        let a_len = a_len.min(VEC_SIZE - a_pos);
        let b_len = b_len.min(VEC_SIZE - b_pos);

        t_del_del(VEC_SIZE, a_pos, a_len, b_pos, b_len)
    }

    #[test]
    fn transform_ins_del(a_pos in 0usize..VEC_SIZE, a_len in 1usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_len in 1usize..VEC_SIZE) {
        let b_len = b_len.min(VEC_SIZE - b_pos);

        t_ins_del(VEC_SIZE, a_pos, a_len, b_pos, b_len)
    }

    #[test]
    fn transform_mov_ins(a_len in 1usize..VEC_SIZE, a_pos in 0usize..VEC_SIZE, a_to in 0usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_len in 1usize..VEC_SIZE) {
        // Limit inputs
        let a_to = filter_to(VEC_SIZE, a_pos, a_to);
        let a_len = a_len.min(VEC_SIZE - a_pos.max(a_to));

        t_mov_ins(VEC_SIZE, a_pos, a_len, a_to, b_pos, b_len)
    }

    #[test]
    fn transform_mov_del(a_pos in 0usize..VEC_SIZE, a_len in 0usize..VEC_SIZE, a_to in 0usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_len in 1usize..VEC_SIZE) {
        // Limit inputs
        let a_to = filter_to(VEC_SIZE, a_pos, a_to);
        let a_len = a_len.min(VEC_SIZE - a_pos.max(a_to));
        let b_len = b_len.min(VEC_SIZE - b_pos);

        t_mov_del(VEC_SIZE, a_pos, a_len, a_to, b_pos, b_len)
    }

    #[test]
    fn transform_mov_mov(a_pos in 0usize..VEC_SIZE, a_to in 0usize..VEC_SIZE, b_pos in 0usize..VEC_SIZE, b_to in 0usize..VEC_SIZE) {
        t_mov_mov(VEC_SIZE, a_pos, a_to, b_pos, b_to)
    }
}
